from typing import Any, Callable, TypeAlias, Protocol

from .channel import Channel
from .connection import Connection
from .message import Message


class ConnectionFactory(Protocol):
    def __call__(self, **kwargs: Any) -> Connection: ...


# Returned by pika.channel.Channel.basic_consume,
# "a tag which may be used to cancel the consumer".
# https://pika.readthedocs.io/en/stable/modules/channel.html
ConsumerTag: TypeAlias = str

OnChannelOpenCallback: TypeAlias = Callable[[Channel], None]
OnMessageCallback: TypeAlias = Callable[[Channel, Message], None]



from pika.exceptions import UnroutableError

from .channel import Channel
from .connection import Connection
from .message import Message
from .message_bus import MessageBus

DEFAULT_MESSAGE_BUS = MessageBus()

blocking_connection = DEFAULT_MESSAGE_BUS.blocking_connection
publisher_connection = DEFAULT_MESSAGE_BUS.publisher_connection

__all__ = [
    "Channel",
    "Connection",
    "DEFAULT_MESSAGE_BUS",
    "Message",
    "MessageBus",
    "UnroutableError",
    "blocking_connection",
    "publisher_connection",
]



import logging

from dataclasses import dataclass, field
from threading import Event, Thread
from typing import Any, Protocol, Self, cast

from hive.common.typing import AnyCallable
from hive.common.units import SECOND

from .channel import Channel
from .connection import Connection

logger = logging.getLogger(__name__)
d = logger.debug


class Invoker(Protocol):
    def __call__(
            self,
            func: AnyCallable,
            *args: Any,
            **kwargs: Any
    ) -> Any: ...


class PublisherConnection(Connection, Thread):
    def __init__(self, *args: Any, **kwargs: Any):
        thread_name = kwargs.pop("thread_name", "Publisher")
        poll_interval = kwargs.pop("poll_interval", 1 * SECOND)
        self._poll_interval = poll_interval.total_seconds()
        Thread.__init__(self, name=thread_name, daemon=True)
        Connection.__init__(self, *args, **kwargs)
        self.is_running = True

    def __enter__(self) -> Self:
        logger.info("Starting publisher thread")
        Thread.start(self)
        return Connection.__enter__(self)

    def run(self) -> None:
        logger.info("%s: thread started", self.name)
        while self.is_running:
            self._pika.process_data_events(time_limit=self._poll_interval)
        logger.info("%s: thread stopping", self.name)
        self._pika.process_data_events(time_limit=self._poll_interval)
        logger.info("%s: thread stopped", self.name)

    def __exit__(self, *exc_info: Any) -> None:
        logger.info("Stopping publisher thread")
        self.is_running = False
        self.join()
        logger.info("Publisher thread stopped")
        return Connection.__exit__(self, *exc_info)

    def _channel(self, *args: Any, **kwargs: Any) -> Channel:
        return cast(Channel, PublisherChannel(
            self._invoke,
            self._invoke(super()._channel, *args, **kwargs),
        ))

    def _invoke(self, func: AnyCallable, *args: Any, **kwargs: Any) -> Any:
        callback = PublisherCallback(func, args, kwargs)
        self._pika.add_callback_threadsafe(callback)
        return callback.join()


@dataclass
class PublisherCallback:
    _func: AnyCallable
    _args: tuple[str]
    _kwargs: dict[str, Any]
    _event: Event = field(default_factory=Event)
    _result: Any = None
    _exception: Exception | None = None

    def __call__(self) -> None:
        d("Entering callback")
        try:
            self._result = self._func(*self._args, **self._kwargs)
        except Exception as exc:
            self._exception = exc
        finally:
            self._event.set()
            del self._func, self._args, self._kwargs
            d("Leaving callback")

    def join(self, *args: Any, **kwargs: Any) -> Any:
        d("Waiting for callback")
        self._event.wait(*args, **kwargs)
        d("Callback returned")
        try:
            if self._exception:
                raise self._exception
            return self._result
        finally:
            del self._result, self._exception


@dataclass
class PublisherChannel:
    _invoker: Invoker
    _channel: Channel

    def __getattr__(self, attr: str) -> Any:
        result = getattr(self._channel, attr)
        if not callable(result):
            return result
        return PublisherInvoker(self._invoker, result)


@dataclass
class PublisherInvoker:
    _invoke: Invoker
    _func: AnyCallable

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._invoke(self._func, *args, **kwargs)



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any, Optional, Self, TYPE_CHECKING

from pika import BlockingConnection as _PikaConnection

from .channel import Channel

if TYPE_CHECKING:
    from .typing import OnChannelOpenCallback


@dataclass
class Connection:
    _pika: _PikaConnection
    _: KW_ONLY
    on_channel_open: Optional[OnChannelOpenCallback] = None

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_info: Any) -> None:
        if self._pika.is_open:
            self._pika.close()

    def _channel(
            self,
            *,
            name: str = "",
            **kwargs: Any
    ) -> Channel:
        return Channel(self._pika.channel(**kwargs), name=name)

    def channel(
            self,
            *,
            name: str = "",
            confirm_delivery: bool = True,
            **kwargs: Any
    ) -> Channel:
        """Like :class:pika.channel.Channel` but with different defaults.

        :param name: Used by `Channel.consume_events()` to construct
             unique queue names.  May be required when more than one
             consumer with the same `Channel.consumer_name` may exist,
             which can happen if processes have multiple channels or
             if multiple processes share the same name.
        :param confirm_delivery: Whether to enable delivery confirmations.
            Hive's default is True.  Use `confirm_delivery=False` for the
            original Pika behaviour.
        """
        channel = self._channel(name=name, **kwargs)
        if confirm_delivery:
            channel._pika.confirm_delivery()  # Don't fail silently.
        if self.on_channel_open:
            self.on_channel_open(channel)
        return channel



from dataclasses import dataclass, field
from typing import Any, Optional, Protocol
from unittest.mock import Mock

import pytest

from cloudevents.abstract import CloudEvent

from .connection import Connection
from .message import Message
from .message_bus import MessageBus


class ConnectionFactory(Protocol):
    def __call__(self, **kwargs: Any) -> Connection: ...


@pytest.fixture
def blocking_connection() -> ConnectionFactory:
    def connect(**kwargs: Any) -> Connection:
        kwargs["connection_attempts"] = 1
        try:
            return MessageBus().blocking_connection(**kwargs)
        except KeyError:
            pytest.skip("Message bus not configured")
        except ConnectionRefusedError as e:
            pytest.skip(f"Message bus not available: {e}")
    return connect


@dataclass(frozen=True)
class ChannelCall:
    method: str
    routing_key: str
    message: Message
    event: Optional[CloudEvent]


@dataclass(frozen=True)
class LogDecoder:
    call_log: list[ChannelCall]


@dataclass(frozen=True)
class MockChannel:
    log_decoder: type[LogDecoder] = LogDecoder
    call_log: list[ChannelCall] = field(default_factory=list)

    def __getattr__(self, attr: str) -> Any:
        return MockMethod(attr, self.call_log)

    @property
    def expect(self) -> LogDecoder:
        return self.log_decoder(self.call_log)


@dataclass(frozen=True)
class MockMethod:
    name: str
    _call_log: list[ChannelCall]

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        if self.name != "basic_publish":
            return
        assert not args
        assert not kwargs.pop("routing_key")
        routing_key = kwargs.pop("exchange")
        mandatory = kwargs.pop("mandatory", False)
        name = "publish_request" if mandatory else "publish_event"
        message = Message(method=Mock(), **kwargs)
        try:
            event = message.event()
        except Exception:
            event = None
        self._call_log.append(ChannelCall(name, routing_key, message, event))
